package com.jdbernard.teammaker

import groovy.beans.Bindable
import groovy.swing.SwingBuilder
import java.awt.GridBagConstraints as GBC
import java.awt.GridBagLayout
import java.awt.BorderLayout
import java.awt.event.MouseEvent
import javax.swing.DefaultListModel
import javax.swing.JFrame
import javax.swing.JOptionPane
import org.apache.log4j.Logger

import static com.jdbernard.teammaker.Game.Team.*

public class ProfessionalPickup {

    /* ======== MODEL ======== */

    public static def swing = new SwingBuilder()
    public static final String version = "0.4"

    // GUI Components
    class View {
        def frame               // main frame
        def team1List           // JList for team 1
        def team2List           // JList for team 2
        def team1WinsButton 
        def team2WinsButton
        def newGameButton
        def sittingList         // JList for sitting players

        def inGamePlayerDropMenu
        def sittingPlayerDropMenu
    }

    View view = new View()

    protected Logger log = Logger.getLogger(getClass())

    // Data components
    class Model {
        @Bindable List team1Players = []
        @Bindable List team2Players = []
        @Bindable List sittingPlayers = []
        @Bindable boolean inGame = false
        @Bindable int teamSize = 4
        @Bindable List popupPlayers = null
        @Bindable int popupTeam;
        @Bindable PlayerChooser playerChooser = new WeightedChooser(hardLimit: 2)
        @Bindable OddsCalculator oddsCalculator = new OddsCalculator()
        @Bindable Properties config
    }

    Model model = new Model()


    /* ======== VIEW ======== */

        /*
           +--------------------------------------------------+
           | JDB Professional Pickup v0.5                _[]X |
           +--------------------------------------------------+
           |                                                  |
           | +-Current-Game--------------+ +-Sitting-Players+ |
           | |                           | |                | |
           | | Team A          Team B    | | Jeptha       3 | |
           | |   Derrick         Eric    | | Jack         2 | |
           | |   Shane           Brandon | | Sam          2 | |
           | |   Shane           Darren  | | Greg         2 | |
           | |   Christian       Brett   | | Frank        1 | |
           | +---------------------------+ | Chris D      1 | |
           |                               | Joshua       1 | |
           |                               |                | |
           |                               |                | |
           |                               |                | |
           |                               |                | |
           |                               |                | |
           |                               +----------------+ |
           +--------------------------------------------------+

         */

    private void initGUI() {
        
        view.inGamePlayerDropMenu = swing.popupMenu(/*'Player Options'*/) {
            menuItem('Bench Player',/* enabled: bind { !model.inGame },*/
                actionPerformed: { benchPlayer(model.popupPlayers,
                    model.popupTeam) })
        }

        view.sittingPlayerDropMenu = swing.popupMenu(/*'Player Options'*/) {
            menuItem('Assign to Team A', /*enabled: bind { !model.inGame },*/
                actionPerformed: {
                    model.popupPlayers.each { assignPlayer(it, 1) }})

            menuItem('Assign to Team B', /*enabled: bind { !model.inGame },*/
                actionPerformed: {
                    model.popupPlayers.each { assignPlayer(it, 2) }})

            menuItem('Set Games Sat',
                actionPerformed: {
                    model.popupPlayers.each { setGamesSat(it) }})
        }

        // TODO
        // settingsDialog = new SettingsDialog(main: this)

        view.frame = swing.frame(title: "JDB Professional Pickup v$version",
            iconImages: [swing.imageIcon("/bb128.png").image,
                swing.imageIcon("/bb64.png").image,
                swing.imageIcon("/bb48.png").image,
                swing.imageIcon("/bb32.png").image,
                swing.imageIcon("/bb16.png").image],
            size: [600, 400], locationRelativeTo: null,
            defaultCloseOperation: JFrame.EXIT_ON_CLOSE) {

            menuBar() {
                menu('Settings') {
                    pointsButtonGroup = buttonGroup()
                    pointsButtonGroup << radioButtonMenuItem("1's and 2's",
                        actionPerformed: { Game.BASE_POINT_VALUE = 1 })
                    pointsButtonGroup << radioButtonMenuItem("2's and 3's",
                        actionPerformed: { Game.BASE_POINT_VALUE = 2 })

                    separator()

                    menuItem('Team Size...',
                        actionPerformed: { setTeamSize() })
                }
            }

            panel() {
                gridBagLayout()

                panel(constraints: gbc(gridx: 0, gridy: 0, fill: GBC.BOTH,
                        insets: [5, 5, 5, 0], weightx: 3, weighty: 1),
                    border: titledBorder(title: 'Current Game')) {
                    
                    gridBagLayout()

                    label('Team A', constraints: gbc(gridx: 0, gridy: 0,
                        fill: GBC.BOTH, insets: [5, 5, 0, 0]))

                    label('Team B', constraints: gbc(gridx: 1, gridy: 0, 
                        fill: GBC.BOTH, insets: [5, 5, 0, 5]))
                    
                    def teamListRenderer = new PlayerListCellRenderer(this,
                        showStats: false, colored: false)

                    view.team1List = list(cellRenderer: teamListRenderer,
                        constraints: gbc(gridx: 0, gridy: 1, fill: GBC.BOTH,
                        insets: [5, 5, 0, 0], weightx: 2, weighty: 2),
                        model: new DefaultListModel(),
                        mouseClicked: { evt ->
                            if (evt.button != MouseEvent.BUTTON3) return
                            showInGamePlayerDropMenu(evt)
                        })

                    view.team1WinsButton = button('Team A Wins',
                        constraints: gbc(gridx: 0, gridy: 2, anchor:
                            GBC.CENTER, insets: [5, 5, 5, 0]),
                        enabled: bind { model.inGame },
                        actionPerformed: { declareLoser(2)})

                    view.team2List = list(cellRenderer: teamListRenderer,
                        constraints: gbc(gridx: 1, gridy: 1, fill: GBC.BOTH,
                        insets: [5, 5, 0, 5], weightx: 2, weighty: 2),
                        model: new DefaultListModel(),
                        mouseClicked: { evt ->
                            if (evt.button != MouseEvent.BUTTON3) return
                            showInGamePlayerDropMenu(evt)
                        })

                    view.team2WinsButton = button('Team B Wins',
                        constraints: gbc(gridx: 1, gridy: 2,
                            anchor: GBC.CENTER, insets: [5, 5, 5, 5]),
                        enabled: bind { model.inGame },
                        actionPerformed: { declareLoser(1)} )

                }

                scrollPane(constraints: gbc(gridx: 1, gridy: 0, gridwidth: 2,
                        fill: GBC.BOTH, insets: [5, 5, 0, 5], weighty: 1),
                    border: titledBorder(title: 'Sitting Players')) {

                    view.sittingList = list(cellRenderer: new PlayerListCellRenderer(
                            this, showStats: true, colored: true),
                            model: new SortedPlayerModel(),
                            mouseClicked: { evt ->
                                if (evt.button != MouseEvent.BUTTON3) return
                                showSittingPlayerDropMenu(evt)
                            })
                }

                button('Next Game', constraints: gbc(gridx: 0, gridy: 1,
                    anchor: GBC.CENTER, insets: [5, 5, 5, 0]),
                    enabled: bind { !model.inGame},
                    actionPerformed: { newGame() })

                button('Add Player', constraints: gbc(gridx: 1, gridy: 1,
                    anchor: GBC.CENTER, insets: [5, 5, 5, 0]),
                    actionPerformed: { addPlayer() })

                view.newGameButton = button('Delete Player',
                    constraints: gbc(gridx: 2, gridy: 1, anchor: GBC.CENTER,
                        insets: [5, 5, 5, 5]),
                    actionPerformed: {
                        model.sittingPlayers.remove(view.sittingList.selectedValues)
                            refreshGUI() })

            }
        }

    }


    

    /* ======== CONTROLLER ======== */

    public static void main(String[] args) {
        def inst = new ProfessionalPickup()
        inst.view.frame.show()
    }

    ProfessionalPickup() {
        model.config = new Properties()

        // look for team maker config file
        File teammakerrc = new File(System.getProperty('user.home'),
            '.teammakerrc')
        if (!teammakerrc.exists())
            getClass().getResourceAsStream('/default-teammakerrc')
            .withInputStream { is -> teammakerrc.text = is.text }

        // load config
        teammakerrc.withInputStream { model.config.load(it) }

        // set configured properties
        Game.BASE_POINT_VALUE = model.config.getProperty('base-point-value',
            '1').toInteger()
        model.teamSize = model.config.getProperty('team.size', '4').toInteger()

        initGUI()
    }

    /**
     * @return the number of players needed to round out even teams.
     */
    public int getSpotsOpen() {
        def spots =  (model.teamSize - model.team1Players.size()) +
                     (model.teamSize - model.team2Players.size())
        return (spots == 0 ? model.teamSize : spots)
    }

    private void setTeamSize() {
        def newSize = JOptionPane.showInputDialog(view.frame,
            "Enter the new team size: ", "Set Team Size...",
            JOptionPane.QUESTION_MESSAGE)

        if (newSize && newSize.isInteger()) model.teamSize = newSize.toInteger()
        else JOptionPane.showMessageDialog(view.frame,
            "$newSize is not a valid team size (it is not a number...).",
            "Set Team Size...", JOptionPane.ERROR_MESSAGE)
    }

    private void addPlayer() {
        def name = JOptionPane.showInputDialog(view.frame,
            "Enter the new player's name: ", "New Player...",
            JOptionPane.QUESTION_MESSAGE)

        if (!name) return

        def player = new Player()
        player.name = name
        player.gamesSat = 0
        model.sittingPlayers << player

        refreshGUI()
    }

    private void newGame() {
        if (model.team1Players.size() < model.teamSize)
            populate(1)

        if (model.team2Players.size() < model.teamSize)
            populate(2)

        recalculateOdds()

        model.inGame = true

    }

    public void declareWinner(Game.Team team) {
        if (team == TEAM1) declareLoser(TEAM2)
        else declareLoser(TEAM1)
    }

    public void declareLoser(Game.Team team) {
        model.sittingPlayers.each { it.gamesSat++ }
        def teamPlayers = playersForTeam(team)
        teamPlayers.each { player ->
            player.gamesSat = 0
            model.sittingPlayers << player
        }

        teamPlayers.clear()

        refreshGUI()

        model.inGame = false
    }

    private void refreshGUI() {

        if (log.isTraceEnabled()) log.trace("Refreshing GUI.")
        if (log.isDebugEnabled()) {
            log.debug("team1Players: ${model.team1Players}")
            log.debug("team2Players: ${model.team2Players}")
        }

        view.team1List.model = new DefaultListModel()
        view.team2List.model = new DefaultListModel()
        view.sittingList.model = new SortedPlayerModel()

        model.team1Players.each { view.team1List.model.addElement(it) }
        model.team2Players.each { view.team2List.model.addElement(it) }
        model.sittingPlayers.each { view.sittingList.model.addElement(it) }

        view.frame.repaint()

        recalculateOdds()

    }

    private void recalculateOdds() {
        def players = model.sittingPlayers
        swing.doOutside {
            model.oddsCalculator.recalculate(players, getSpotsOpen(), model.playerChooser)
            swing.edt { view.sittingList.repaint() }
        }

    }

    private void populate(Game.Team team) {
        def teamPlayers = playersForTeam(team)
        while (teamPlayers.size() < model.teamSize) {
            def player = model.playerChooser.choose(model.sittingPlayers)
            teamPlayers << player
            model.sittingPlayers.remove(player)
        }

        refreshGUI();
    }

    private benchPlayer(Player p, Game.Team team) {
        def teamPlayers = playersForTeam(team)
        teamPlayers.remove(p)
        model.sittingPlayers << p
        refreshGUI()
    }

    private assignPlayer(Player p, Game.Team team) {
        def teamPlayers = playersForTeam(team)
        model.sittingPlayers.remove(p)
        teamPlayers << p
        refreshGUI()
    }

    private void showSittingPlayerDropMenu(def evt) {
        int index = view.sittingList.locationToIndex(evt.point)
        if (view.sittingList.isSelectedIndex(index))
            this.model.popupPlayers = view.sittingList.selectedValues
        else {
            view.sittingList.selectedIndex = index
            this.model.popupPlayers = [view.sittingList.model.getElementAt(index)]
        }

        view.sittingPlayerDropMenu.show(view.sittingList, evt.point.@x, evt.point.@y)
    }

    private void showInGamePlayerDropMenu(def evt) {
        def teamList = evt.source
        int index = teamList.locationToIndex(evt.point)
        if (teamList.isSelectedIndex(index)) 
            this.model.popupPlayers = teamList.selectedValues 
        else {
            teamList.selectedIndex = index
            model.popupPlayers = [teamList.model.getElementAt(index)]
        }

        this.model.popupTeam = (teamList == view.team1List ? 1 : 2)

        view.inGamePlayerDropMenu.show(teamList, evt.point.@x, evt.point.@y)
    }

    private void setGamesSat(Player p) {
        String value = JOptionPane.showInputDialog(view.frame,
            "How many games has ${p.name} sat?", "Set Games Sat.",
            JOptionPane.QUESTION_MESSAGE)
        int newGames

        if (!value) return
        try { newGames = value.toInteger() }
        catch (Throwable t) {
            JOptionPane.showMessageDialog(view.frame,
                "The value $value is not a valid number.",
                "Invalid Input", JOptionPane.ERROR_MESSAGE)
            setGamesSat(p)
            return
        }
        
        p.gamesSat = newGames

        refreshGUI()
    }

    private List playersForTeam(Game.Team team) {
        if (team == Game.Team.TEAM1) return model.team1Players
        else return model.team2Players
    }

    private List listForTeam(Game.Team team) {
        if (team == Game.Team.TEAM11) return view.team1List
        else return view.team2List
    }
}
